// Copyright (C) 2023  Benedikt Zumtobel
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package pwchadm

import (
	"context"
	"database/sql"
	"encoding/hex"
	"errors"
	"fmt"
	"log"
	"net/mail"
	"os"
	"os/exec"
	"runtime/debug"
	"strings"

	"golang.org/x/crypto/bcrypt"
	"golang.org/x/crypto/sha3"
	"golang.org/x/term"
	"gopkg.in/yaml.v3"
)

type Config struct {
	DB struct {
		Host     string `yaml:"host"`
		DBName   string `yaml:"db_name"`
		User     string `yaml:"user"`
		Password string `yaml:"password"`
		SSLMode  string `yaml:"ssl_mode"`
	} `yaml:"db"`
	Bcrypt struct {
		Cost int `yaml:"cost"`
	} `yaml:"bcrypt"`
}

func connectToDatabase(cfg *Config) (*sql.DB, error) {
	if cfg == nil {
		return nil, errors.New("config is nil")
	}

	connStr := "user=" + cfg.DB.User + " password=" + cfg.DB.Password +
		" dbname=" + cfg.DB.DBName + " host=" + cfg.DB.Host +
		" sslmode=" + cfg.DB.SSLMode

	db, err := sql.Open("postgres", connStr)
	if err != nil {
		return nil, err
	}

	return db, nil
}

// Executes the sql schema against the database
// This is the minimum requirement for the database.
func InitializeDatabase(cfg *Config, sqlSchema string) error {
	db, err := connectToDatabase(cfg)
	if err != nil {
		log.Print(err)
		log.Print("ERROR: cannot connect to database")
	}
	defer db.Close()

	_, err = db.Exec(sqlSchema)

	return err
}

// Writes a new user into the database with the given values.
// The password gets bcrypt hashed
// The domain is created on first occurrence.
func createUserInDatabase(ctx context.Context, cfg *Config, tx *sql.Tx, username, domain, password string, quota int) error {
	domainPresent, err := domainExists(ctx, tx, domain)
	if err != nil {
		return err
	}

	if !domainPresent {
		domainErr := createDomainInDatabase(ctx, tx, domain)
		if domainErr != nil {
			return domainErr
		}
	}

	passwordHash, err := hashPassword(password, cfg)
	if err != nil {
		return err
	}

	_, err = tx.ExecContext(ctx, `
    INSERT INTO accounts (
        username,
        domain,
        password,
        mail_crypt_salt,
        quota,
        enabled,
        sendonly
    ) VALUES (
        $1, $2, $3,
        ENCODE(gen_random_bytes(32), 'hex'),
        $4, true, false
    );
`, username, domain, passwordHash, quota)

	if err != nil {
		return err
	}

	return nil
}

func createDomainInDatabase(ctx context.Context, tx *sql.Tx, domain string) error {
	_, err := tx.ExecContext(ctx, "INSERT INTO domains (domain) VALUES ($1);", domain)

	return err
}

func domainExists(ctx context.Context, tx *sql.Tx, domain string) (bool, error) {
	rows, err := tx.QueryContext(ctx, "SELECT domain FROM domains WHERE domain = $1;", domain)
	if err != nil {
		return false, err
	}

	if rows.Err() != nil {
		return false, err
	}

	defer rows.Close()

	if rows.Next() {
		return true, nil
	}

	return false, nil
}

// This function runs the doveadm command to encrypt a user's mailbox.
func encryptMailbox(db *sql.DB, username, domain, password string) error {
	// Get the salt from the database
	var salt string
	err := db.QueryRow("SELECT mail_crypt_salt FROM accounts WHERE username = $1 AND domain = $2;", username, domain).Scan(&salt)
	if err != nil {
		return err
	}

	// generate the password being used to encrypt the user's private key
	privatePassword := sha3.Sum512([]byte(salt + password))
	privatePasswordString := hex.EncodeToString(privatePassword[:])
	email := username + "@" + domain
	pluginString := "plugin/mail_crypt_private_password=" + privatePasswordString

	// run doveadm to encrypt the mailbox
	cmd := exec.Command("/bin/doveadm", "-o", pluginString, "mailbox", "cryptokey", "generate", "-u", email, "-UR")
	err = cmd.Run()
	if err != nil {
		return err
	}

	return nil
}

func hashPassword(password string, cfg *Config) (string, error) {
	bytes, err := bcrypt.GenerateFromPassword([]byte(password), cfg.Bcrypt.Cost)

	return string(bytes), err
}

func IsValidEmail(email string) bool {
	_, err := mail.ParseAddress(email)

	return err == nil
}

// This function either enables or disables a user in the database.
func ChangeUserStatus(email string, enable bool, cfg *Config) bool {
	components := strings.Split(email, "@")
	username, domain := components[0], components[1]

	db, err := connectToDatabase(cfg)
	if err != nil {
		log.Print(err)
		log.Print("ERROR: cannot connect to database")
	}
	defer db.Close()

	result, err := db.Exec("UPDATE accounts SET enabled = $1 WHERE username = $2 AND domain = $3;", enable, username, domain)
	if err != nil {
		log.Print(err)

		return false
	}

	affected, _ := result.RowsAffected()

	return affected == 1
}

func promptForEmail() (string, error) {
	var email string
	fmt.Print("Email: ") //nolint:forbidigo
	_, err := fmt.Scanf("%s", &email)

	return email, err
}

func promptForPassword(confirm bool) (string, error) {
	var prompt string
	if confirm {
		prompt = "\nConfirm password: "
	} else {
		prompt = "Password: "
	}

	fmt.Print(prompt) //nolint:forbidigo
	bytePassword, err := term.ReadPassword(int(os.Stdin.Fd()))
	if err != nil {
		return "", err
	}

	return string(bytePassword), nil
}

func AddUser(cfg *Config) error {
	email, err := promptForEmail()
	if err != nil {
		log.Fatal(err)
	}

	if !IsValidEmail(email) {
		log.Fatal("invalid email format")
	}

	password, err := promptForPassword(false)
	if err != nil {
		log.Fatal(err)
	}

	confirmPassword, err := promptForPassword(true)
	if err != nil {
		log.Fatal(err)
	}

	fmt.Println() //nolint:forbidigo

	// Check if passwords match
	if password != confirmPassword {
		log.Fatal("passwords do not match")
	}

	// Prompt user for storage quota
	var quota int

	fmt.Print("Storage quota in MB (min 1): ") //nolint:forbidigo
	_, err = fmt.Scanf("%d", &quota)
	if err != nil {
		log.Fatal(err)
	}

	// Extract username and domain
	components := strings.Split(email, "@")
	if len(components) != 2 {
		log.Panicln("invalid email format")
	}
	username, domain := components[0], components[1]

	db, err := connectToDatabase(cfg)
	if err != nil {
		log.Print(err)
		log.Print("ERROR: cannot connect to database")
	}
	defer db.Close()

	// Create SQL transaction
	ctx := context.Background()
	tx, err := db.BeginTx(ctx, nil)
	if err != nil {
		return err
	}
	defer tx.Rollback() //nolint:errcheck

	// Create user
	err = createUserInDatabase(ctx, cfg, tx, username, domain, password, quota)
	if err != nil {
		log.Println("SQL transaction canceled")

		return err
	}

	// Commit SQL transaction
	err = tx.Commit()
	if err != nil {
		return err
	}

	// Encrypt mailbox
	err = encryptMailbox(db, username, domain, password)
	if err != nil {
		log.Println("cannot encrypt mailbox")

		return err
	}

	fmt.Println("User added and mailbox encrypted") //nolint:forbidigo

	return nil
}

// reads config file.
func ReadFile(cfg *Config, configPath string) error {
	file, err := os.Open(configPath)
	if err != nil {
		return err
	}

	decoder := yaml.NewDecoder(file)
	err = decoder.Decode(cfg)
	if err != nil {
		return err
	}

	return nil
}

func PrintBuildInfo(version string) {
	buildInfo, ok := debug.ReadBuildInfo()
	if !ok {
		log.Fatal("Can't read BuildInfo")
	}

	fmt.Println("Version:")       //nolint:forbidigo
	fmt.Printf("  %s\n", version) //nolint:forbidigo

	fmt.Println("Built with:")                //nolint:forbidigo
	fmt.Printf("  %s\n", buildInfo.GoVersion) //nolint:forbidigo
}
